"use strict"

/*
  This audio code comes from Hover-Pixel 384, and some of it predates that.
  It wasn't originally written with the intention of appearing in an 
  open-source project.
*/

const AUDIO_TIMING_LATENCY=1/60;
const ENGINE_NOISE_CROSSFADE_TIME=1/8;
const ENGINE_NOISE_STOP_TIME=1/8;
const ENGINE_NOISE_LOOP_POINT=2;

// weird 19-TET sfx system, borrowed from hyper pixel ten thousand
// [tone offset,base36 tone, base17 volume,wave parameter,slot]
// tone/volume define curves at 60 data points per second
// wave parameter is basically how many different PRNG steps to use in the
// waveform before going back to the first one
// slots exist to determine when a new sound cancels a previous one
const SFX_TONE_ZERO=264*Math.SQRT2/44100;
const SFX_VOLUME_SCALE=.5;
const SFX_DATA={
 bounce:[33,"bcba",
            "2111",2,'bounce'],
 damage:[14,"bbccbbccbbccbb",
	    'a9876543222111',93,'damage'],
 lastdamage:[14,"bbccb9876543210",
	        'a98765444444333',24,'damage'],
 death:[7,"bbbccccbbbaaa999877655532210",
  	  'aaaaaa9887665444444444333333',24,'damage'],
 gameOver:[-12,"bbbbcccccbbbbaaaa99998777655554322100",
	       'aaaaaa9998887777666555444444444333333',24,'damage'], 
 coin:[52,"66hip",
          "21x21",4,"pickup"],
 bigcoin:[52,"6666abbbbijjjjj",
             "444434433333332",4,"pickup"],
 life: [19,"0368behjmprux","8887776665432",4,'bonus'],
 fuel: [40,'bbbbijjjjoppppppo',
     	   'cba9cba98ba987642',10,'pickup'],
 armor: [40,'jjjjoppppijjjjjji',
	    'cba9cba98ba987642',10,'pickup'],
 checkpoint: [82,'009bbhjjj','444332211',8,'pickup'],
 warning: [32,'qqqqqqqomk',
	      '2333333321',3,'warning'] 
}


function renderSFX(name,data,audioContext) {
 // this turns the strings into numbers and passes them to
 // makeSoundEffectForParams
 if(data.length==5) {
  var toneOffset=+data[0];
  var toneStr=data[1].trim();
  var volStr=data[2].trim();
  var period=+data[3];
  var channel=data[4].trim();
  var length=Math.min(volStr.length,toneStr.length);
  var tones=new Array(length);
  for(var j=0;j<length;++j) {
   var c=toneStr[j];
   if(c>='0' && c<='9') { tones[j]=+c+toneOffset; }
   else if(c>='A' && c<='Z') { tones[j]=c.charCodeAt(0)-65+10+toneOffset}
   else if(c>='a' && c<='z') { tones[j]=c.charCodeAt(0)-97+10+toneOffset}
  }
  var volumes=new Array(length);
  for(var j=0;j<length;++j) {
   var c=volStr[j];
   if(c>='0' && c<='9') { volumes[j]=+c/16; }
   else if(c>='A' && c<='G') { volumes[j]=(c.charCodeAt(0)-65+10)/16 }
   else if(c>='a' && c<='g') { volumes[j]=(c.charCodeAt(0)-97+10)/16 }
   else if(c=='x' || c=="X") { volumes[j]=1/32; }
   else if(c=='y' || c=="Y") { volumes[j]=1/64; }
   else if(c=='z' || c=="Z") { volumes[j]=1/128; }
  }
  var buf=makeSoundEffectForParams(period,tones,volumes,audioContext);
  return [buf,channel]
 }
}

function makeSoundEffectForParams(period,tones,volumes,audioContext) {
 // quirky wave generator; the waveform is made using a linear congruential
 // PRNG that gets reset periodically
 var osc=0;
 var pos=0;
 var prng=1;
 var stepCount=0;
 var waveAmplitude=1;
 var buf=audioContext.createBuffer(1,
				   735*tones.length,
				   44100);
 var data=buf.getChannelData(0);
 for(var i=0;i<tones.length;++i) {
  var vol=volumes[i];
  var tone=tones[i];
  var freq=SFX_TONE_ZERO*Math.pow(2,tone/19);
  for(var j=0;j<735;++j) {
   data[pos]=waveAmplitude*vol*SFX_VOLUME_SCALE;
   pos+=1;
   osc=osc+freq;
   if(osc>=2) {
    console.log("frequency overflow warning");
   }
   while(osc>=1) {
    osc-=1;
    ++stepCount;
    if(stepCount==period) {
     prng=1;
     waveAmplitude=1;
     stepCount=0;
    }
    else {
     prng=(prng*48271)%2147483647;
     waveAmplitude=1-(prng%257)/128;
    }
   }
  }
 }
 return buf;
}

/*
  Class that keeps track of audio state, wrapping an AudioContext
*/
class AudioManager {
 constructor() {
  this.audioContext=new AudioContext();
  this.slots={}
  this.sfxBuffers={} 
  this.sfxSlotNames={}
  for(var k in SFX_DATA) {
   var pair=renderSFX(k,SFX_DATA[k],this.audioContext);
   this.sfxBuffers[k]=pair[0];
   this.sfxSlotNames[k]=pair[1];
  }
  this.engineBuffers={}
  this._makeEngineNoises();
  this.currentEngineNoise="0,0";  
  this._initGraph();
 }
 onAudioEnablingInput() {
  if(this.audioContext.state=="suspended") {
   this.audioContext.resume();
  }
 }
 pause() {
  var pauseTime=this.audioContext.currentTime+AUDIO_TIMING_LATENCY;
  if(!this.paused) {
   for(var k in this.slots) {
    this.slots[k].pause(pauseTime);
   }
   // if a sound was fading out and we abandoned its source,
   // we can't pause it, and the only way to disconnect it involves
   // ignoring AUDIO_TIMING_LATENCY which is bad here, so pausing
   // might not be an instantaneous cut to audio zero in all cases.
   this.paused=true;
  }
 }
 unpause() {
  var unpauseTime=this.audioContext.currentTime+AUDIO_TIMING_LATENCY;
  if(this.paused) {
   for(var k in this.slots) {
    this.slots[k].unpause(unpauseTime);
   }
   this.paused=false;
  }
 }
 stopEngineNoise() {
  this._stopSlot("engine",
		 this.audioContext.currentTime+AUDIO_TIMING_LATENCY,
		 ENGINE_NOISE_STOP_TIME);
  this.currentEngineNoise="0,0";
 }
 startSFX(name) {
  if(this.sfxBuffers[name]) {
   this._startBuffer(this.sfxBuffers[name],
		     this.sfxSlotNames[name],
		     null,
		     0
		    );
  }
  else {
   console.log("No SFX named "+name);
  }
 }
 setEngineNoise(xDir,yDir) {
  var key=xDir+","+yDir;
  if(key!=this.currentEngineNoise) {
   this.currentEngineNoise=key;
   var buffer=this.engineBuffers[key];
   this._startBuffer(buffer,"engine",
		     ENGINE_NOISE_LOOP_POINT,
		     ENGINE_NOISE_CROSSFADE_TIME);
   if(!buffer && (xDir || yDir)) {
    console.log("No engine noise buffer for key "+key);
   }   
  }
 }
 stopAll(fadeOutTime) {
  if(fadeOutTime==undefined) {
   for(var k in this.slots) {
    this._stopSlot(k,
		   this.audioContext.currentTime,
		   fadeOutTime);
   }
   // we no longer have a pointer to nodes that were already fading out,
   // but we can disconnect them like this
   this.currentEngineNoise="0,0"
   this._destroyGraph();
   this._initGraph();
  }
  else {
   for(var k in this.slots) {
    this._stopSlot(k,
		   this.audioContext.currentTime+AUDIO_TIMING_LATENCY,
		   fadeOutTime);
   }
  }
 }
 _startBuffer(buffer,slot,loopStartOrNull,fadeLength) {
  var startTime=this.audioContext.currentTime+AUDIO_TIMING_LATENCY;
  this._stopSlot(slot,startTime,fadeLength);
  if(buffer) {
   this.slots[slot]=new ManagedSource(this,buffer,startTime,loopStartOrNull);
  }
 }
 _stopSlot(slot,fadeStart,fadeLength) {
  if(slot in this.slots) {
   this.slots[slot].stop(fadeStart,fadeLength);
   delete this.slots[slot];
  }
 }
 _destroyGraph() {
  if(this.clipper) {
   this.clipper.disconnect();
   delete this.clipper;
  }
 }
 _initGraph() {
  this.clipper=this.audioContext.createWaveShaper();
  this.clipper.curve=new Float32Array([-1,1]);
  this.clipper.connect(this.audioContext.destination);
 }
 _makeEngineNoises() {
  var base=381;
  this.engineBuffers["0,-1"]=
   this._makeEngineNoise(ENGINE_NOISE_CROSSFADE_TIME,
			 2/6,2/6,
			 base+2,base,
			 12345);
  this.engineBuffers["1,-1"]=
   this._makeEngineNoise(ENGINE_NOISE_CROSSFADE_TIME,
			 Math.sqrt(5)/6,Math.sqrt(3)/6,
			 base+1,base*5/6,2,34567);
  
  this.engineBuffers["1,0"]=
   this._makeEngineNoise(ENGINE_NOISE_CROSSFADE_TIME,
			 Math.sqrt(6)/6,Math.sqrt(2)/6,
			 base,base*3/4,45678);
  this.engineBuffers["1,1"]=
   this._makeEngineNoise(ENGINE_NOISE_CROSSFADE_TIME,
			 Math.sqrt(5)/6,Math.sqrt(3)/6,
			 base-1,base*2/3,56789);

  this.engineBuffers["-1,-1"]=
   this._makeEngineNoise(ENGINE_NOISE_CROSSFADE_TIME,
			 Math.sqrt(3)/6,Math.sqrt(5)/6,
			 base+1,base*5/6,2,345670);
  
  this.engineBuffers["-1,0"]=
   this._makeEngineNoise(ENGINE_NOISE_CROSSFADE_TIME,
			 Math.sqrt(2)/6,Math.sqrt(6)/6,
			 base,base*3/4,456780);
  this.engineBuffers["-1,1"]=
   this._makeEngineNoise(ENGINE_NOISE_CROSSFADE_TIME,
			 Math.sqrt(3)/6,Math.sqrt(5)/6,
			 base-1,base*2/3,567890);
  
  this.engineBuffers["0,1"]=
   this._makeEngineNoise(ENGINE_NOISE_CROSSFADE_TIME,
			 2/6,2/6,
			 base-2,base/2,
			 23456);
 }
 _makeEngineNoise(attack,leftGain,rightGain,noiseHz,lowpass,seed) {
  // prng noise enveloped and run through a lowpass filter;
  // details were arrived at empirically during Hover-Pixel 384 development
  // and I have forgetten if they ever had any psychoacoustic justification
  var length=24000*5;
  var buffer=this.audioContext.createBuffer(2,
					    length,
					    24000);
  var left=buffer.getChannelData(0);
  var right=buffer.getChannelData(1);
  var mult=16807, mod=Math.pow(2,31)-1;
  var lpAccumulator=0;
  var hpAccumulator=0;
  var lpAlpha=lowpass/24000/(lowpass/24000+.5/Math.PI);
  var osc=0;
  var loopPoint=Math.floor(24000*ENGINE_NOISE_LOOP_POINT);
  var loopLevel;
  var lastLoop=false
  var loopThresh=length-Math.ceil(24000/noiseHz);
  for(var i=0;i<length;++i) {
   var vol;
   if(i<attack*24000) { vol=i/(attack*24000); } else { vol=1; }
   var level;
   if(lastLoop) {
    // I think this was something I did so the loop from the end of the
    // buffer back to the loop point wouldn't have an audible
    // non-lowpass-filtered popping sound
    level=loopLevel;
   }
   else {
    var level=(seed>>8)%8;
    level=(level*2/7-1)*vol;
   }
   lpAccumulator=level*lpAlpha+lpAccumulator*(1-lpAlpha);
   left[i]=lpAccumulator*leftGain;
   right[i]=lpAccumulator*rightGain;
   osc+=noiseHz;
   if(osc>=24000) {
    osc=osc%24000;
    seed=(seed*mult)%mod;
    if(i>=loopThresh) { lastLoop=true; }
   }
   if(i==loopPoint) { loopLevel=lpAccumulator; }
  }
  return buffer;
 }
}

/*
  Class that keeps track of one sound's state, wrapping an 
  AudioBufferSourceNode and some other nodes connected to it. The AudioContext
  the node is ultimately outputting to is in an AudioManager.
*/
class ManagedSource {
 constructor(manager,buffer,startTime,loopStartOrNull) {
  this.bufferPositionAtStartTime=0;
  this.startTime=startTime;
  this.buffer=buffer;
  this.manager=manager;
  if(loopStartOrNull!=null) {
   this.loopStart=loopStartOrNull;
   this.loop=true;
  }
  else {
   this.loop=false;
  }
  this._makeNodes();
  this.bufferSource.start(startTime);
 }
 _makeNodes() {
  this.bufferSource=this.manager.audioContext.createBufferSource();
  this.bufferSource.buffer=this.buffer;
  if(this.loop) {
   this.bufferSource.loop=true;
   this.bufferSource.loopStart=this.loopStart;
   this.bufferSource.loopEnd=this.buffer.duration;
  }
  else {
   this.bufferSource.loop=false;
  }
  this.gain=this.manager.audioContext.createGain();
  this.gain.gain.value=1;
  this.bufferSource.connect(this.gain);
  this.gain.connect(this.manager.clipper);
 }
 pause(pauseTime) {
  if(!this.paused) {
   this.bufferPositionAtPauseTime=
    this.bufferPositionAtStartTime+(pauseTime-this.startTime)
   if(this.loop) {
    var timeInLoop=this.bufferPositionAtPauseTime-this.loopStart;
    if(timeInLoop>0) {
     var loopLength=this.buffer.duration-this.loopStart;
     this.bufferPositionAtPauseTime=this.loopStart+(timeInLoop%loopLength);
    }
   }
   this.bufferSource.stop(pauseTime);
   this.paused=true;
  }
 }
 unpause(unpauseTime) {
  if(this.paused) {
   this._makeNodes();
   this.startTime=unpauseTime
   this.bufferPositionAtStartTime=this.bufferPositionAtPauseTime;
   this.bufferSource.start(unpauseTime,this.bufferPositionAtPauseTime);
   this.paused=false;
  }
 }
 stop(fadeStart,fadeLength) {
  if(fadeLength) {
   this.gain.gain.setValueAtTime(1,fadeStart);
   this.gain.gain.linearRampToValueAtTime(0,fadeStart+fadeLength)
   this.bufferSource.stop(fadeStart+fadeLength);
  }
  else {
   this.bufferSource.stop(fadeStart);
  }
 }
}

