"use strict"

/* How big's the subwindow we're drawing the level to? 
   If this is any larger than MIN_PNG_WIDTH and MIN_PNG_HEIGHT, there needs
   to be special-case code for levels that don't fill the subwindow. */
const VIEWPORT_WIDTH=288;
const VIEWPORT_HEIGHT=202;

/* Where is the player within that subwindow (before considering 
   edge-of-map scrolling thresholds)? */
const PLAYER_CAMERA_X_OFFSET=144;
const PLAYER_CAMERA_Y_OFFSET=100;

/* Where is that subwindow within the game canvas? */
const VIEWPORT_LEFT=96;
const VIEWPORT_TOP=0;

/* How big's the game canvas itself? */
const GAME_CANVAS_WIDTH=384;
const GAME_CANVAS_HEIGHT=216;

/* How big's the area containing the game canvas, with a little padding? */
const PADDED_GAME_CANVAS_WIDTH=400;
const PADDED_GAME_CANVAS_HEIGHT=232;

// How long does the background flash on damage, and what color?
const DAMAGE_FLASH_FRAMES=10;
const DAMAGE_FLASH_COLOR="#400000";

// how long does "toast" UI text stay onscreen for, and how slowly
// does it move upward?
const TOAST_EXPIRATION_FRAMES=90;
 const TOAST_FRAMES_PER_PIXEL=8;

/* The dynamically-sized canvas on the page and its 2D context,
   and the containing element that has its border padding */
var screenCanvas, screenContext, screenCanvasArea;

/* The fixed-sized canvas for drawing the game at 1:1 pixel scale */
var gameCanvas, gameContext;

/* A viewport-sized canvas used for compositing ImageData objects */
var viewportScratchCanvas, viewportScratchContext;

/* The spritesheet image element */
var spritesheet

/* Draw function; only called via requestAnimationFrame. Responsible for
   all the visual output of the running game, including canvas size management.
 */
function drawGame() {

 // clear the requestAnimationFrame handle and numUndrawnEngineTicks so
 // the engine knows we've done a draw
 animationFrameRequestHandle=null;
 numUndrawnEngineTicks=0;

 // make sure we have a canvas to draw to
 setUpCanvasState();

 if(currentLevel.atWinScreen || currentLevel.atLoseScreen) {
  gameContext.drawImage(spritesheet,512,512,
			GAME_CANVAS_WIDTH,GAME_CANVAS_HEIGHT,
			0,0,
			GAME_CANVAS_WIDTH,GAME_CANVAS_HEIGHT);
  gameContext.save();
  try {
   gameContext.translate(100,0);
   gameContext.drawImage(spritesheet,
			 256,(currentLevel.atWinScreen?464:448),
			 112,10,
			 65,12,
			 112,10);   
   var flightTimeY=32;
   if(currentLevel.coinsSpawned>0) {
    drawCoinLine(32);
    flightTimeY=40;
   }
   drawFlightTimeLine(40);
   if(currentLevel.endScreenFrameNumber>=END_SCREEN_WAIT_FRAMES) {
    gameContext.drawImage(spritesheet,
			  256,480,
			  112,13,
			  65,81,
			  112,13);
   }
  }
  finally {
   gameContext.restore();
  }
 }
 else {
  // draw background and stats
  gameContext.drawImage(spritesheet,0,512,GAME_CANVAS_WIDTH,GAME_CANVAS_HEIGHT,
			0,0,GAME_CANVAS_WIDTH,GAME_CANVAS_HEIGHT);
  // draw subwindow where the game is happening -- this uses canvas clipping,
  // so it's wrapped in save/restore to avoid accidentally making the
  // clip permanent
  gameContext.save();
  try {
   gameContext.beginPath();
   gameContext.rect(VIEWPORT_LEFT,VIEWPORT_TOP,VIEWPORT_WIDTH,
		    VIEWPORT_HEIGHT);
   gameContext.clip();
   gameContext.translate(VIEWPORT_LEFT,VIEWPORT_TOP);
   drawGameViewport();
  }
  finally {
   gameContext.restore();
  }
 }

 // either way, draw stats...
 drawWhiteNumeral2(currentLevel.player.lives,2, 24, 102);
 drawWhiteNumeral2(currentLevel.player.coins,3, 60, 102);
 drawWhiteNumeral2(currentLevel.player.maxCoins,3, 76, 102);
 drawWhiteNumeral2(currentLevel.player.fuel,4, 112, 102);
 drawWhiteNumeral2(currentLevel.player.maxFuel,4, 132, 102);
 drawWhiteNumeral2(currentLevel.player.armor, 1, 179, 102);
 drawWhiteNumeral2(currentLevel.player.maxArmor, 1, 187, 102);

 screenContext.drawImage(gameCanvas,0,0,GAME_CANVAS_WIDTH,GAME_CANVAS_HEIGHT,
			 0,0,screenCanvas.width,screenCanvas.height);
}

function setUpCanvasState() {
 if(!screenCanvas) {
  screenCanvas=document.getElementById("hyopCanvas");
  screenContext=screenCanvas.getContext("2d");
  screenContext.imageSmoothingEnabled=false
  screenCanvasArea=document.getElementById("hyopCanvasArea");
  gameCanvas=document.createElement("canvas");
  gameCanvas.width=GAME_CANVAS_WIDTH;
  gameCanvas.height=GAME_CANVAS_HEIGHT;
  gameContext=gameCanvas.getContext("2d");
  gameContext.imageSmoothingEnabled=false
  viewportScratchCanvas=document.createElement("canvas");
  viewportScratchCanvas.width=VIEWPORT_WIDTH;
  viewportScratchCanvas.height=VIEWPORT_HEIGHT;
  viewportScratchContext=viewportScratchCanvas.getContext("2d");
  viewportScratchContext.imageSmoothingEnabled=false
  spritesheet=document.getElementById("spritesheet");
  var firstTimeSizing=true;
 }
 else {
  var firstTimeSizing=false;
 }
 var widths=Math.floor(screenCanvasArea.clientWidth/PADDED_GAME_CANVAS_WIDTH);
 var heights=Math.floor(
  screenCanvasArea.clientHeight/PADDED_GAME_CANVAS_HEIGHT
 );
 var scale=Math.max(1,Math.min(widths,heights));
 if(firstTimeSizing || screenCanvas.width!=scale*GAME_CANVAS_WIDTH) {
  screenCanvas.width=scale*GAME_CANVAS_WIDTH;
  screenCanvas.height=scale*GAME_CANVAS_HEIGHT;
  screenContext.imageSmoothingEnabled=false
  screenCanvas.style["margin-left"]=(-screenCanvas.width/2)+"px";
  screenCanvas.style["margin-top"]=(-screenCanvas.height/2)+"px";
 }
}

function drawGameViewport() {
 // going into this, gameContext is clipped and transformed
 // so it's just the viewport

 // background is normally black, but we may want a flash..
 if(currentLevel.player.invulnerabilityFrames>
    DAMAGE_INVULNERABILITY_FRAMES-DAMAGE_FLASH_FRAMES)
 {
  gameContext.fillStyle=DAMAGE_FLASH_COLOR;
  gameContext.fillRect(0,0,VIEWPORT_WIDTH,VIEWPORT_HEIGHT);
 }

 // where's the camera
 var cameraLeft=currentLevel.player.x-PLAYER_CAMERA_X_OFFSET;
 var cameraTop=currentLevel.player.y-PLAYER_CAMERA_Y_OFFSET;
 cameraLeft=Math.max(0,cameraLeft);
 cameraTop=Math.max(0,cameraTop); 
 cameraLeft=Math.min(cameraLeft,currentLevel.width-VIEWPORT_WIDTH);
 cameraTop=Math.min(cameraTop,currentLevel.height-VIEWPORT_HEIGHT);
 
 // collect visible sprites, stratified by layer
 var firstSpriteLayer=0, lastSpriteLayer=0;
 var spriteLayers={}
 currentLevel.sprites.forEach( (s)=>{
  if(s.visibleForCamera(cameraLeft,cameraTop)) {
   var layer=Math.floor(s.layer || 0);
   firstSpriteLayer=Math.min(layer,firstSpriteLayer);
   lastSpriteLayer=Math.max(layer,lastSpriteLayer);
   if(!(layer in spriteLayers)) { spriteLayers[layer]=[]; }
   spriteLayers[layer].push(s);
  }
 });
 
 // draw behind-background sprites
 for(var layer=firstSpriteLayer; layer<0; ++layer) {
  if(layer in spriteLayers) {
   spriteLayers[layer].forEach((s) => {
    s.drawForCamera(cameraLeft, cameraTop);
   })
  }  
 }

 drawBackgroundTilesForCamera(cameraLeft, cameraTop);
 
 // draw in-front-of-background sprites
 for(var layer=0; layer<=lastSpriteLayer; ++layer) {
  if(layer in spriteLayers) {
   spriteLayers[layer].forEach((s) => {
    s.drawForCamera(cameraLeft, cameraTop);
   })
  }  
 }

 if(currentLevel.player.visibleForCamera(cameraLeft,cameraTop)) {
  currentLevel.player.drawForCamera(cameraLeft, cameraTop);
 }

 drawToast(cameraLeft, cameraTop);
}


/* Draw background tiles into the game context (which, when this is called,
   is already clipped and transformed to be just about the viewport). 
   This is a little complicated because the tiles are ImageData objects and
   we want their alpha channels to work in case of behind-the-solid-layer
   objects. */
function drawBackgroundTilesForCamera(cameraLeft, cameraTop) {
 // we're using viewportScratchContext because
 // putImageData would erase behind-the-background sprites if we used it
 // directly.
 viewportScratchContext.clearRect(0,0,VIEWPORT_WIDTH,VIEWPORT_HEIGHT);

 // which tile has the top-left pixel of the viewport, clamped to actual tiles?
 var firstTileXIndex=Math.max(0,Math.floor(cameraLeft/SCROLL_TILE_SIZE));
 var firstTileYIndex=Math.max(0,Math.floor(cameraTop/SCROLL_TILE_SIZE));
 // which tile has the bottom-right pixel of the viewport, clamping Y to
 // actual tiles?
 var lastTileXIndex=Math.floor((cameraLeft+VIEWPORT_WIDTH-1)/SCROLL_TILE_SIZE);
 var lastTileYIndex=Math.min(
  currentLevel.tiles.length-1,
  Math.floor((cameraTop+VIEWPORT_HEIGHT-1)/SCROLL_TILE_SIZE));

 // iterate all those tiles (and clamp X to actual tiles), put them to
 // viewport scratch
 for(var tileYIndex=Math.max(0,firstTileYIndex);
     tileYIndex<=lastTileYIndex;
     ++tileYIndex) {
  for(var tileXIndex=firstTileXIndex;
      (tileXIndex<=lastTileXIndex &&
       tileXIndex<currentLevel.tiles[tileYIndex].length);
      ++tileXIndex) {
   viewportScratchContext.putImageData(
    currentLevel.tiles[tileYIndex][tileXIndex],
    tileXIndex*SCROLL_TILE_SIZE-cameraLeft,
    tileYIndex*SCROLL_TILE_SIZE-cameraTop)
  }
 }
 // now draw viewport scratch to game context; this step will apply the
 // alpha from the tiles.
 gameContext.drawImage(viewportScratchCanvas,0,0);
}

function drawCoinLine(top) {
 var coinDigits=(""+currentLevel.coinsSpawned).length;
 gameContext.drawImage(spritesheet,
		       16,472,44,5,
		       85,top,44,5);
 drawYellowNumeral(currentLevel.player.totalCoins,
		    coinDigits,
		   135,top);
 gameContext.drawImage(spritesheet,
		       16,456,3,5,
		       135+coinDigits*4,top,3,5);
 drawYellowNumeral(currentLevel.coinsSpawned,
		   coinDigits,
		   139+coinDigits*4,top);
}

function drawFlightTimeLine(top) {
 var seconds=Math.floor(currentLevel.player.flightTime/60);
 var justSeconds=seconds%60;
 var minutes=Math.floor(seconds/60)
 if(minutes>99) { minutes=99; justSeconds=59; }
 gameContext.drawImage(spritesheet,
		       16,464,45,5,
		       85,top,45,5);
 drawYellowNumeral(minutes, 2,
		   135,top);
 gameContext.drawImage(spritesheet,
		       20,456,3,5,
		       143,top,3,5);
 drawYellowNumeral(justSeconds, 2, 
		   147,top);
}

const TOAST_GRAPHICS={
 "MAX FUEL UP":{left:256,top:408,width:47,height:7},
 "MAX ARMOR UP":{left:256,top:416,width:53,height:7},
 "EXTRA LIFE":{left:256,top:424,width:41,height:7},
 "LOW ARMOR":{left:256,top:432,width:41,height:7},
 "LOW FUEL":{left:256,top:440,width:35,height:7},
}

function drawToast(cameraLeft, cameraTop) {
 var toast=currentLevel.currentToast;
 if(toast) {
  var age=currentLevel.frameNumber-toast.startFrame;
  if(age<TOAST_EXPIRATION_FRAMES) {
   var movePixels=Math.floor(age/TOAST_FRAMES_PER_PIXEL);
   var rect=TOAST_GRAPHICS[toast.name];
   var left=currentLevel.player.x-cameraLeft-Math.floor(rect.width/2);
   var top=currentLevel.player.y-cameraTop-rect.height-movePixels;
   left=Math.max(0,left);
   left=Math.min(VIEWPORT_WIDTH-rect.width,left);
   gameContext.drawImage(spritesheet,
			 rect.left,rect.top,rect.width,rect.height,
			 left,top,rect.width,rect.height);
  }
 }
}

function drawNumeral(value,numDigits,left,top, texTop,scale) {
 scale=scale||1;
 var strValue=""+value;
 while(strValue.length<numDigits) { strValue="0"+strValue; }
 if(strValue.length>numDigits) {
  strValue="9";
  while(strValue.length<numDigits) { strValue="9"+strValue; }
 }
 for(var pos=0;pos<strValue.length;++pos) {
  var digit=strValue.charCodeAt(pos)-48;
  if(digit>=0 && digit<=9) {
   gameContext.drawImage(spritesheet,
			 16+digit*4*scale, texTop, 3*scale, 5*scale,
			 left+pos*4*scale, top, 3*scale, 5*scale);
  }
 }   
}

function drawWhiteNumeral(value,numDigits,left,top) {
 drawNumeral(value,numDigits,left,top,480);
}

function drawWhiteNumeral2(value,numDigits,left,top) {
 drawNumeral(value,numDigits,left*2,top*2,320,2);
}

function drawYellowNumeral(value,numDigits,left,top) {
 drawNumeral(value,numDigits,left,top,448);
}
