"use strict"

// How many pixels are there per subpixel? This is really just used by the
// PlayerSprite; other objects have individual speedDenominator values
// that manage their subpixels
const SUBPIXEL_SCALE=512;

// How hard does the player thrust, in subpixels?
const DIAGONAL_THRUST=17; // per axis, so net thrust is this times sqrt2
const CARDINAL_THRUST=24; // along one axis

// What speed is the player clamped to?
const MAX_VELOCITY=1024;

// If the player bounces gently off a wall, what's the slowest speed they can
// bounce at?
const MIN_BOUNCE_VELOCITY=256;

// How long is the timeout between taking consecutive damage, with and
// without having any fuel left?
const DAMAGE_INVULNERABILITY_FRAMES=120;
const DAMAGE_INVULNERABILITY_FRAMES_NO_FUEL=4;

// How long is the timeout after leaving a checkpoint or other parked state
// before damage can be taken?
const UNPARK_INVULNERABILITY_FRAMES=45;

// How long do we hold on the end screen before offering new-game-start?
const END_SCREEN_WAIT_FRAMES=45;

// the main game logic setInterval handle goes here
var engineIntervalHandle;

// when are we expecting the next engine tick, in performance.now() units
var engineNextTickTime;

// how many engine ticks have happened since last animation draw
var numUndrawnEngineTicks;

// the data structure of the current level, including its sprites and player
var currentLevel;

// the requestAnimationFrame handle for drawing the game state goes here
var animationFrameRequestHandle;

// track some keyboard stuff
var keyState;

// instance of AudioManager class that's defined in audio.js
var audioManager;

/* Initialize the game state for the level */
function startGame(level) {
 keyState={held:{},fresh:{},freshText:{}}
 window.onkeydown=function(e) {
  if(!audioManager) {
   audioManager=new AudioManager();
  }
  else {
   audioManager.onAudioEnablingInput();
  }
  if(!e.shiftKey && !e.ctrlKey) { e.preventDefault(); }
  if(!keyState.held[e.code]) {   
   keyState.held[e.code]=true;
   keyState.fresh[e.code]=true;
   keyState.freshText[e.key]=true;
  }
 }
 window.onkeyup=function(e) {
  if(!e.shiftKey && !e.ctrlKey) { e.preventDefault(); }
  if(keyState.held[e.code]) {   
   delete keyState.held[e.code];
  }
 }
 window.onblur=function(e) {
  keyState.held={}
  keyState.fresh={}
  keyState.freshText={}
 }
 // level came from loadLevel in level_loader.js; possibly it has also
 // gotten other properties set since then and we're restarting it
 level.frameNumber=0;
 level.sprites=[]
 level.coinsSpawned=0;
 level.hasExit=false;
 level.stamps.asList.forEach((stamp)=> {
  var regStamp= stamp.registeredStamp;
  if(regStamp.spriteConstructor) {
   var ctor=regStamp.spriteConstructor;
   var sprite=new ctor(stamp.left,stamp.top);
   level.sprites.push(sprite);
   if(sprite.isCoin) { level.coinsSpawned+=sprite.coinCount; }
   if(sprite.isExit) { level.hasExit=true; }
  }
 });
 currentLevel=level;
 level.player=new PlayerSprite(level.startX,level.startY);
 level.atWinScreen=false;
 level.atLoseScreen=false;
 setToast(null);
 if(!engineIntervalHandle) {
  engineNextTickTime=performance.now()+1000/TICKS_PER_SECOND;
  engineIntervalHandle=setInterval(engineTickCheck,250/TICKS_PER_SECOND);
  numUndrawnEngineTicks=0;
  if(!animationFrameRequestHandle) {
   animationFrameRequestHandle=requestAnimationFrame(drawGame);
  }
 }
}

function exitToLoaderScreen() {
 if(engineIntervalHandle) {
  clearInterval(engineIntervalHandle);
  engineIntervalHandle=undefined;
 }
 awaitFile();
}

/* the loopy part of the game logic loop; run as many ticks as the timer
   and draw state indicate */
function engineTickCheck() {
 if(performance.now()>
    engineNextTickTime+MAX_TICKS_PER_DRAW*1000/TICKS_PER_SECOND) {
  // the next tick is much too far in the past; jump the clock ahead
  // so we don't have more than one max draw's worth of catch-up ticks waiting
  engineNextTickTime=
   performance.now()-MAX_TICKS_PER_DRAW*1000/TICKS_PER_SECOND;
 }
 if(numUndrawnEngineTicks>=MAX_TICKS_PER_DRAW &&
    performance.now()>engineNextTickTime) {
  // we're still waiting for an overdue draw; update the engine next tick time
  // even though we haven't actually ticked
  engineNextTickTime=performance.now()
  return;
 }
 while(numUndrawnEngineTicks<MAX_TICKS_PER_DRAW &&
       performance.now()>=engineNextTickTime) {
  runEngineTick();
  keyState.fresh={};
  keyState.freshText={};
  ++numUndrawnEngineTicks;
  if(!animationFrameRequestHandle) {
   animationFrameRequestHandle=requestAnimationFrame(drawGame);
  }
  engineNextTickTime+=1000/TICKS_PER_SECOND;
 }
}

function runEngineTick() {
 if(currentLevel.atWinScreen || currentLevel.atLoseScreen) {
  currentLevel.endScreenFrameNumber+=1;
  if(currentLevel.endScreenFrameNumber>=END_SCREEN_WAIT_FRAMES) {
   var newGame, restartGame;
   if(keyState.freshText["R"] || keyState.freshText["r"]) {
    restartGame=true;
   }
   else if(keyState.fresh["KeyR"]) { restartGame=true; }
   if(newGame) {
    exitToLoaderScreen();
   }
   else if(restartGame) {
    startGame(currentLevel);
   }
  }
 }
 else {  
  currentLevel.frameNumber+=1;
  for(var i=0;i<currentLevel.sprites.length;++i) {
   currentLevel.sprites[i].runFrame();
  }
  var controlState=interpretKeyState();
  currentLevel.player.runFrame(controlState);
  currentLevel.sprites=currentLevel.sprites.filter( (s)=>!s.despawned );
  if(currentLevel.player.reachedExit) {
   startWinScreen();
  }
  else if(!currentLevel.hasExit && currentLevel.coinsSpawned>0) {
   if(currentLevel.player.totalCoins>=currentLevel.coinsSpawned) {
    startWinScreen();
   }
  }
 }
}

function interpretKeyState(){
 var controlState={x:0,y:0,freshInput:false};

 if(keyState.fresh["ArrowRight"]) {
  controlState.x+=1;
  controlState.freshInput=true;
 }
 else if(keyState.held["ArrowRight"]) {
  controlState.x+=1;
 }

 if(keyState.fresh["ArrowLeft"]) {
  controlState.x-=1;
  controlState.freshInput=true;
 }
 else if(keyState.held["ArrowLeft"]) {
  controlState.x-=1;
 }

 if(keyState.fresh["ArrowUp"]) {
  controlState.y-=1;
  controlState.freshInput=true;
 }
 else if(keyState.held["ArrowUp"]) {
  controlState.y-=1;
 }

 if(keyState.fresh["ArrowDown"]) {
  controlState.y+=1;
  controlState.freshInput=true;
 }
 else if(keyState.held["ArrowDown"]) {
  controlState.y+=1;
 }
 
 return controlState;
}

/* Player sprite is special in various ways. Among other things,
   it's not part of currentLevel.sprites, it doesn't have a numbered sprite
   layer, it has a variety of animation states to consider ,and its collision
   is automatically one pixel with no mask shape. 
   It coexists with the HYOPSprite base class via duck-typing, not inheritance
   (and since it isn't called in the main sprite loops, it doesn't need
   to have much in common)
*/
class PlayerSprite {
 constructor(x,y) {
  this.spawnX=x;
  this.spawnY=y;
  this.x=x;
  this.y=y;
  this.state="parked"
  this.subX=0;
  this.subY=0;
  this.xv=0;
  this.yv=0;
  // gravity is linear and can go negative
  this.gravity=3+currentLevel.gravityModifier;
  // fuel has a linear high end and nonlinear low end; the starting fuel
  // is always five of the upgrade increment
  if(currentLevel.fuelModifier>0) {
   this.fuelIncrement=100*(2+currentLevel.fuelModifier);
  }
  else {
   switch(currentLevel.fuelModifier) {
   case 0: this.fuelIncrement=200; break;
   case -1: this.fuelIncrement=100; break;
   case -2: this.fuelIncrement=50; break;
   case -3: this.fuelIncrement=25; break;
   case -4: this.fuelIncrement=10; break;
   case -5: this.fuelIncrement=5; break;
   case -6: this.fuelIncrement=2; break;
   default: this.fuelIncrement=1; break;    
   }
  }
  this.maxFuel=this.fuelIncrement*5;
  // armor is linear but caps at 0
  this.maxArmor=Math.max(0,3+currentLevel.armorModifier);
  // lives are linear but cap at 0
  this.lives=Math.max(0,5+currentLevel.livesModifier);

  this.coins=0;
  this.maxCoins=currentLevel.coinsSpawned;
  this.totalCoins=0;
  this.invulnerabilityFrames=0;
  this.flightTime=0;
  

  this.armorIncrement=1;
  this.armor=this.maxArmor;
  this.fuel=this.maxFuel;
  
  for(var i in currentLevel.sprites) {
   var s=currentLevel.sprites[i];
   if(s.isCheckpoint &&
      this.x>=s.left &&
      this.y>=s.top &&
      this.x<s.left+s.width &&
      this.y<s.top+s.height) {
    this.checkpoint=s;
    s.onPlayerSpawnHere();
   }   
  }
 }
 runFrame(controlState) {
  switch(this.state) {
  case "parking": this.runParkingFrame(controlState); break;
  case "parked": this.runParkedFrame(controlState); break;   
  case "flight": this.runFlightFrame(controlState); break;
  }
 }
 runParkingFrame(controlState) {
  var targetX=Math.floor(this.parkTarget.left+this.parkTarget.width/2);
  var targetY=Math.floor(this.parkTarget.top+this.parkTarget.height/2);
  if(this.x<targetX) { this.x+=1; }
  if(this.y<targetY) { this.y+=1; }
  if(this.x>targetX) { this.x-=1; }
  if(this.y>targetY) { this.y-=1; }
  if(this.x==targetX && this.y==targetY) {
   this.state="parked";
   setToast(null);
   this.parkTarget.onPlayerPark();
   if(!this.reachedExit) {
    this.fuel=this.maxFuel;
    this.armor=this.maxArmor;
   }
  }
 }
 runParkedFrame(controlState) {  
  if(controlState.freshInput) {
   this.state="flight";
   this.invulnerabilityFrames=UNPARK_INVULNERABILITY_FRAMES;
   this.runFlightFrame(controlState);
  }
 }
 runFlightFrame(controlState) {
  ++this.flightTime;
  if(this.invulnerabilityFrames>0) { --this.invulnerabilityFrames; }
  this.applyThrust(controlState);
  this.applyGravity();
  this.applyVelocityLimit();
  this.applyVelocity();
  if(this.getLowFuelWarningState().beep) {
   audioManager.startSFX("warning");
  }  
  // need to test collision even if not moving this frame, because
  // an enemy could have moved onto the stationary player
  var collided=this.checkCollisions(0,0);
  if(!collided) {
   this.resolveSubpixelOverflow();
  }
 }
 applyThrust(controlState) {
  if(this.fuel>0) {
   if(controlState.x && controlState.y) {
    this.xv+=controlState.x*DIAGONAL_THRUST;
    this.yv+=controlState.y*DIAGONAL_THRUST;
    --this.fuel;
    audioManager.setEngineNoise(controlState.x,controlState.y);
   }
   else if(controlState.x || controlState.y) {
    this.xv+=controlState.x*CARDINAL_THRUST;
    this.yv+=controlState.y*CARDINAL_THRUST;
    --this.fuel;
    if(this.getLowFuelWarningState().toast) {
     setToast("LOW FUEL");
    }
    audioManager.setEngineNoise(controlState.x,controlState.y);
   }
   else {
    audioManager.stopEngineNoise();
   }
  }
  else {
   audioManager.stopEngineNoise();
  }
 }
 applyGravity() {
  this.yv+=this.gravity;
 }
 applyVelocityLimit() {
  var velocitySquared=this.xv*this.xv+this.yv*this.yv;
  if(velocitySquared>MAX_VELOCITY*MAX_VELOCITY) {
   var ceilVelocity=Math.ceil(Math.sqrt(velocitySquared));
   this.xv=Math.trunc(this.xv*MAX_VELOCITY/ceilVelocity);
   this.yv=Math.trunc(this.yv*MAX_VELOCITY/ceilVelocity);   
  }
 }
 applyVelocity() {
  this.subX+=this.xv;
  this.subY+=this.yv;
 }
 resolveSubpixelOverflow() {
  // spend subpixels to move step by step until out of pixels to move or
  // loop is broken
  // loop interleaves X and Y steps, not that it's likely to be noticeable
  // in play whether it's stair-stepping or not
  while(Math.abs(this.subX)>=SUBPIXEL_SCALE ||
	Math.abs(this.subY)>=SUBPIXEL_SCALE) {
   if(this.subX>=SUBPIXEL_SCALE) {
    this.subX-=SUBPIXEL_SCALE;
    this.x++;
    var collided=this.checkCollisions(1,0);
    if(collided) { break; }
   }
   if(this.subX<=-SUBPIXEL_SCALE) {
    this.subX+=SUBPIXEL_SCALE;
    this.x--;
    var collided=this.checkCollisions(-1,0);
    if(collided) { break; }
   }
   if(this.subY>=SUBPIXEL_SCALE) {
    this.subY-=SUBPIXEL_SCALE;
    this.y++;
    var collided=this.checkCollisions(0,1);
    if(collided) { break; }
   }
   if(this.subY<=-SUBPIXEL_SCALE) {
    this.subY+=SUBPIXEL_SCALE;
    this.y--;
    var collided=this.checkCollisions(0,-1);
    if(collided) { break; }
   }      
  }
  // in case the loop was broken with entire pixels of movement left,
  // cap off any remaining subpixel overflow
  if(this.subX>=SUBPIXEL_SCALE) { this.subX=SUBPIXEL_SCALE-1; }
  if(this.subY>=SUBPIXEL_SCALE) { this.subY=SUBPIXEL_SCALE-1; }
  if(this.subX<=-SUBPIXEL_SCALE) { this.subX=-SUBPIXEL_SCALE+1; }
  if(this.subY<=-SUBPIXEL_SCALE) { this.subY=-SUBPIXEL_SCALE+1; }
 }
 checkCollisions(xDir,yDir) {
  // check if the player is now colliding; if so, bounce back a pixel
  // in the opposite of the given direction and return true
  if(this.x<0 || this.y<0 ||
     this.x>=currentLevel.width || this.y>=currentLevel.height ||
     currentLevel.pixelData[ (this.y*currentLevel.width+this.x)*4+3]) {
   // collision with world map: out of bounds,
   // or on a level pixel with nonzero alpha
   this.xv=Math.trunc(this.xv/2);
   this.yv=Math.trunc(this.yv/2);
   this.bounceAway(xDir, yDir);
   audioManager.startSFX("bounce");
   if(!this.invulnerabilityFrames) {
    // wall bounces regardless of invulnerability, so no extra damage bounce
    this.takeDamage(false);
   }
   return true;
  }
  for(var i=0;i<currentLevel.sprites.length; ++i) {
   var sprite=currentLevel.sprites[i];
   if(!sprite.despawned &&
      sprite.testPlayerCollision(this.x,this.y)) {
    var collided = sprite.onPlayerTouch();
    if(collided) { return true; }
   }
  }  
  // no collision, clear to continue movement this frame
  return false; 
 }
 bounceAway(xDir, yDir) {
  this.x-=xDir;
  this.y-=yDir;
  if(xDir>0) {
   this.subX=0;
   if(this.xv>0) { this.xv=-this.xv; }
   if(this.xv>-MIN_BOUNCE_VELOCITY) { this.xv=-MIN_BOUNCE_VELOCITY; }
  }
  if(xDir<0) {
   this.subX=0;
   if(this.xv<0) { this.xv=-this.xv; }
   if(this.xv<MIN_BOUNCE_VELOCITY) { this.xv=MIN_BOUNCE_VELOCITY; }
  }
  if(yDir>0) {
   this.subY=0;
   if(this.yv>0) { this.yv=-this.yv; }
   if(this.yv>-MIN_BOUNCE_VELOCITY) { this.yv=-MIN_BOUNCE_VELOCITY; }
  }
  if(yDir<0) {
   this.subY=0;
   if(this.yv<0) { this.yv=-this.yv; }
   if(this.yv<MIN_BOUNCE_VELOCITY) { this.yv=MIN_BOUNCE_VELOCITY; }
  }
 }
 takeDamage(andBounce) {
  if(this.armor<=0) {
   this.loseLife();
   return;
  }
  --this.armor;
  if(this.armor==0) {
   setToast("LOW ARMOR");
   audioManager.startSFX("lastdamage");
  }
  else {
   audioManager.startSFX("damage");
  }
  if(this.fuel>0) {
   this.invulnerabilityFrames=DAMAGE_INVULNERABILITY_FRAMES;
  }
  else {
   this.invulnerabilityFrames=DAMAGE_INVULNERABILITY_FRAMES_NO_FUEL;
  }
  if(andBounce) {
   this.xv=-Math.trunc(this.xv/4);
   this.yv=-Math.trunc(this.yv/4);
  }
 }
 loseLife() {
  if(this.lives<=0) {
   audioManager.startSFX("gameOver");
   startLoseScreen();
   return;
  }
  setToast(null);
  audioManager.startSFX("death");
  this.lives-=1;
  this.xv=0; this.yv=0;
  this.subX=0; this.subY=0;
  this.armor=this.maxArmor;
  this.fuel=this.maxFuel;
  this.state="parked";
  audioManager.stopEngineNoise();
  if(this.checkpoint) {
   this.x=this.checkpoint.left+Math.floor(this.checkpoint.width/2);
   this.y=this.checkpoint.top+Math.floor(this.checkpoint.height/2);
   this.checkpoint.onPlayerSpawnHere();
  }
  else {
   this.x=this.spawnX;
   this.y=this.spawnY;
  }
 }
 visibleForCamera(cameraLeft, cameraTop) {
  return !this.gameOver;
 }
 drawForCamera(cameraLeft, cameraTop) {
  var whichFrame;
  if(this.invulnerabilityFrames%8>0 &&
     this.invulnerabilityFrames%8<5) {
   whichFrame=2;
  }
  else if(this.getLowFuelWarningState().flash) {
   whichFrame=3;
  }
  else if(this.armor==0) {
   whichFrame=1;
  }
  else {
   whichFrame=0;
  }
  gameContext.drawImage(spritesheet,
			0+whichFrame*2, 480, 1, 1,
			this.x-cameraLeft, this.y-cameraTop, 1, 1); 
 }
 getLowFuelWarningState() {
  var slowWarningLevel=Math.ceil(Math.min(this.maxFuel/2,200));
  if(this.fuel>slowWarningLevel) {
   // ample fuel, no warning
   return {flash:false, beep:false, toast: false};
  }
  if(this.fuel<=0) {
   // out of fuel, stay in flash color but don't beep any more
   return {flash:true, beep:false, toast: false};
  }
  if(this.fuel>100) {
   // slow warning, every 30 frames
   var cycle=currentLevel.frameNumber%30;
  }
  else {
   // fast warning, every 15 frames
   var cycle= currentLevel.frameNumber%15;
  }
  // toast is only used when this is called immediately after spending a
  // fuel point; if that brings fuel to exactly the warning level, then
  // we pop up the text on that frame
  return {flash:cycle<5, beep:cycle==0, toast: this.fuel==slowWarningLevel};
 }
 startParking(target) {
  audioManager.stopEngineNoise();
  this.state="parking";
  this.parkTarget=target;
  this.subX=0; this.subY=0;
  this.xv=0;
  this.yv=0;
  this.invulnerabilityFrames=0;
 }
 arriveAtCheckpoint(checkpoint) {
  this.checkpoint=checkpoint;
  currentLevel.sprites.forEach( (s)=> {
   if("onPlayerReachedCheckpoint" in s) {
    s.onPlayerReachedCheckpoint();
   }
  });
 }
 grantItem(type) {
  switch(type) {
  case "fuel":
   this.maxFuel+=this.fuelIncrement;
   audioManager.startSFX("fuel");
   setToast("MAX FUEL UP");
   break;
  case "armor":
   this.maxArmor+=this.armorIncrement;
   audioManager.startSFX("armor");
   setToast("MAX ARMOR UP");
   break;
  case "1up":
   ++this.lives;
   audioManager.startSFX("life");
   setToast("EXTRA LIFE");
   break;
  case "exit": this.reachedExit=true; break;
  }
 }
 grantCoins(count) {
  this.totalCoins+=count;
  this.coins+=count;
  var gotLife=false;
  if(gotLife) {
   audioManager.startSFX("life");
   setToast("EXTRA LIFE");
  }
  else if(count>1) {
   audioManager.startSFX("bigcoin");
  }
  else if(count==1) {
   audioManager.startSFX("coin");
  }
 }
}

function setToast(name) {
 if(name) {
  currentLevel.currentToast={name:name, startFrame:currentLevel.frameNumber}
 }
 else {
  currentLevel.currentToast=null;
 }
}

function startWinScreen()  {
 audioManager.stopEngineNoise();
 currentLevel.atWinScreen=true;
 currentLevel.endScreenFrameNumber=0;
}

function startLoseScreen()  {
 audioManager.stopEngineNoise();
 currentLevel.atLoseScreen=true;
 currentLevel.endScreenFrameNumber=0;
}


class NoncustomSprite {
 // a lot of things in common with CustomSprite, but kept separate so
 // modding CustomSprite won't break it
 constructor() {
  this.layer=0; // just in front of terrain, behind NoncustomSprites
 }
 setRect(left,top,width,height) {
  this.left=left|0
  this.top=top|0
  this.width=width|0
  this.height=height|0
 }
 setAnimation(spritesheetLeft, spritesheetTop, spritesheetGrid,
	      animationFrameCount, gameFramesPerAnimationFrame) {
  this.spritesheetLeft=spritesheetLeft|0;
  this.spritesheetTop=spritesheetTop|0;
  this.spritesheetGrid=spritesheetGrid|0;
  this.animationFrameCount=animationFrameCount|0;
  this.gameFramesPerAnimationFrame=gameFramesPerAnimationFrame|0;
  this.animationCycleCounter=0;
  this.texLeft=this.spritesheetLeft;
  this.texTop=this.spritesheetTop;
 }
 visibleForCamera(cameraLeft, cameraTop) {
  return this.left+this.width>cameraLeft &&
   this.top+this.height>cameraTop &&
   this.left<cameraLeft+VIEWPORT_WIDTH &&
   this.top<cameraTop+VIEWPORT_HEIGHT &&
   this.texLeft!=undefined;
 }
 drawForCamera(cameraLeft,cameraTop) {
  if(this.texLeft!=undefined) {
   gameContext.drawImage(spritesheet,
			 this.texLeft, this.texTop,
			 this.width, this.height,
			 this.left-cameraLeft, this.top-cameraTop,
			 this.width, this.height);
  }
 }
 runFrame() {
  // subclasses overriding this should call the superclass!
  ++this.animationCycleCounter;
  var whichFrame = Math.floor(
   this.animationCycleCounter/this.gameFramesPerAnimationFrame
  );
  whichFrame=whichFrame%this.animationFrameCount;
  this.texLeft=this.spritesheetLeft+this.spritesheetGrid*whichFrame;
 }
 testPlayerCollision(x,y) {
  if(x<this.left || y<this.top ||
     x>=this.left+this.width || y>=this.top+this.height ||
     this.texLeft==undefined) {
   return false;
  }
  var texX=x-this.left+this.texLeft;
  var texY=y-this.top+this.texTop;
  if(texX<0 || texY<0 ||
     texX>=SPRITESHEET_MASK.width || texY>=SPRITESHEET_MASK.height) {
   return false;
  }
  var texIndex=texY*SPRITESHEET_MASK.width+texX;
  return SPRITESHEET_MASK.alpha[texIndex]>0;
 }
}

class ParkableSprite extends NoncustomSprite {
 onPlayerTouch() {
  if(this.isTouchable()) {
   currentLevel.player.startParking(this);
   return true;
  }
  return false;
 }
 isTouchable() {
  // checkpoints can make this false
  return true;
 }
 onPlayerPark() { } // every sprite has its own thing to do here
}

class CheckpointSprite extends ParkableSprite {
 constructor(x,y) {
  super();
  this.setRect(x,y,11,11);
  this.setAnimation(64,496,16,2,32);
  this.isCheckpoint=true;
 }
 onPlayerSpawnHere() {
  this.recentPlayerOverlap=true;
  this.spritesheetTop=480;
  this.texTop=this.spritesheetTop;
 }
 onPlayerPark() {
  audioManager.startSFX("checkpoint");
  this.recentPlayerOverlap=true;
  currentLevel.player.arriveAtCheckpoint(this);
 }
 runFrame() {
  super.runFrame();
  if(this.recentPlayerOverlap &&
     !this.testPlayerCollision(currentLevel.player.x,
			       currentLevel.player.y)) {
   this.recentPlayerOverlap=false;
  }
  this.spritesheetTop=(currentLevel.player.checkpoint==this)?480:496;
  this.texTop=this.spritesheetTop;
 }
 isTouchable() {
  return !this.recentPlayerOverlap;
 }
 static stamp=[
  "...#####...",
  "..#.....#..",
  ".#.......#.",
  "#...###...#",
  "#..#...#..#",
  "#..#......#",
  "#..#...#..#",
  "#...###...#",
  ".#.......#.",
  "..#.....#..",
  "...#####...",
 ]

 static startStamp=[
 "...#####...",
  "..#.....#..",
  ".#.......#.",
  "#...###...#",
  "#..#......#",
  "#..#.#....#",
  "#..#......#",
  "#...###...#",
  ".#.......#.",
  "..#.....#..",
  "...#####...",
 ]
}

registerStamp(CheckpointSprite.stamp,CheckpointSprite,"checkpoint");
registerStamp(CheckpointSprite.startStamp,CheckpointSprite,"startcheckpoint");

class FuelSprite extends ParkableSprite {
 constructor(x,y) {
  super();
  this.setRect(x,y,13,13);
  this.setAnimation(64,464,16,4,4);
 }
 onPlayerPark() {
  this.despawned=true;
  currentLevel.player.grantItem("fuel");
 }
 static stamp=[
  "....#####....",
  "..#########..",
  ".#.##...##.#.",
  ".#.#######.#.",
  "#..#.....#..#",
  "#..#.###.#..#",
  "#..#.#...#..#",
  "#..#.###.#..#",
  "#..#.#...#..#",
  ".#.#.#...#.#.",
  ".#..#...#..#.",
  "..##.###.##..",
  "....#####....",    
 ]
}

registerSprite(FuelSprite)


class ArmorSprite extends ParkableSprite {
 constructor(x,y) {
  super();
  this.setRect(x,y,13,13);
  this.setAnimation(64,448,16,4,4);
 }
 onPlayerPark() {
  this.despawned=true;
  currentLevel.player.grantItem("armor");
 }
 static stamp=[
  "....#####....",
  "..##.....##..",
  ".#..#####..#.",
  ".#.#.....#.#.",
  "#..#..#..#..#",
  "#..#.#.#.#..#",
  "#..#.###.#..#",
  "#..#.#.#.#..#",
  "#..#.#.#.#..#",
  ".#..#...#..#.",
  ".#...#.#...#.",
  "..##..#..##..",
  "....#####....",    
 ]
}

registerSprite(ArmorSprite)

class OneUpSprite extends ParkableSprite {
 constructor(x,y) {
  super();
  this.setRect(x,y,13,13);
  this.setAnimation(64,432,16,2,8);
 }
 onPlayerPark() {
  this.despawned=true;
  currentLevel.player.grantItem("1up");
 }
 static stamp=[
  "....#####....",
  "..##.....##..",
  ".#.........#.",
  ".#.........#.",
  "#.#..#.#.####",
  "###..#.#.#.##",
  "#.#..#.#.####",
  "#.#..#.#.#..#",
  "####.###.#..#",
  ".#.........#.",
  ".#.........#.",
  "..##.....##..",
  "....#####....",    
 ]
}

registerSprite(OneUpSprite)

class BigCoinSprite extends ParkableSprite {
 constructor(x,y,count,spritesheetTop) {
  super();
  this.setRect(x,y,13,13);
  this.setAnimation(64,spritesheetTop,16,4,2);
  this.isCoin=true;
  this.coinCount=count;
 }
 onPlayerPark() {
  this.despawned=true;
  currentLevel.player.grantCoins(this.coinCount);
 }
}

class TenCoinSprite extends BigCoinSprite {
 constructor(x,y) {
  super(x,y,10,384);
 }
 static stamp=[
  "....#####....",
  "..##.....##..",
  ".#.........#.",
  ".#.........#.",
  "#..#...###..#",
  "#.##...#.#..#",
  "#..#...#.#..#",
  "#..#...#.#..#",
  "#.###..###..#",
  ".#.........#.",
  ".#.........#.",
  "..##.....##..",
  "....#####....",
 ]
}

registerSprite(TenCoinSprite);

class TwentyFiveCoinSprite extends BigCoinSprite {
 constructor(x,y) {
  super(x,y,25,400);
 }
 static stamp=[
  "....#####....",
  "..##.....##..",
  ".#.........#.",
  ".#.........#.",
  "#.###..###..#",
  "#...#..#....#",
  "#.###..###..#",
  "#.#......#..#",
  "#.###..###..#",
  ".#.........#.",
  ".#.........#.",
  "..##.....##..",
  "....#####....",
 ]
}

registerSprite(TwentyFiveCoinSprite);


class FiftyCoinSprite extends BigCoinSprite {
 constructor(x,y) {
  super(x,y,50,416);
 }
 static stamp=[
  "....#####....",
  "..##.....##..",
  ".#.........#.",
  ".#.........#.",
  "#.###..###..#",
  "#.#....#.#..#",
  "#.###..#.#..#",
  "#...#..#.#..#",
  "#.###..###..#",
  ".#.........#.",
  ".#.........#.",
  "..##.....##..",
  "....#####....",
 ]
}

registerSprite(FiftyCoinSprite);

class OneCoinSprite extends NoncustomSprite {
 constructor(x,y) {
  super();
  this.setRect(x,y,7,7);
  this.setAnimation(64,368,16,4,2);
  this.isCoin=true;
  this.coinCount=1;
 }
 onPlayerTouch() {
  this.despawned=true;
  currentLevel.player.grantCoins(this.coinCount);
  return false; // not a "collision" for momentum computation purposes!
 }
 static stamp=[
  "..###..",
  ".#...#.",
  "#.....#",
  "#.....#",
  "#.....#",
  ".#...#.",
  "..###..",
 ]
}

registerSprite(OneCoinSprite);

class ExitSprite extends ParkableSprite {
 constructor(x,y) {
  super();
  this.setRect(x,y,13,13);
  this.setAnimation(64,352,16,4,4);
  this.isExit=true; // this flag stops the game from ending at the last coin
 }
 onPlayerPark() {
  this.despawned=true;
  currentLevel.player.grantItem("exit");
 }
 static stamp=[
  "....#####....",
  "..##.....##..",
  ".#...###...#.",
  ".#.##...##.#.",
  "#..#.....#..#",
  "#.#...#...#.#",
  "#.#..#.#..#.#",
  "#.#...#...#.#",
  "#..#.....#..#",
  ".#.##...##.#.",
  ".#...###...#.",
  "..##.....##..",
  "....#####....",
 ]
}

registerSprite(ExitSprite);
