"use strict"

// png.js throws an out of memory error on big files; at 8000
// the browser will complain about a slow script but it won't necessarily
// fail to load
const MAX_PNG_SIZE=8000;

// The level should be big enough to fill the viewport, to avoid
// needing any special-case camera scrolling logic
const MIN_PNG_WIDTH=144;
const MIN_PNG_HEIGHT=108;

// How many gameplay frames per second to try to run; visual FPS might be lower
// than this but never higher
const TICKS_PER_SECOND=60;

// How many gameplay frames it's okay to batch together between visual
// draw frames; after this many gameplay frames, the game won't attempt
// any more game logic until a draw happens
const MAX_TICKS_PER_DRAW=5;

// if a file input happens when this is false, ignore it
var isFileExpected;

/* called once at page load */
function initHYOP() {
 unpackCollisionMask();
 awaitFile();
}

function initHPGU2() {
 document.getElementById("hyopCanvasArea").hidden=true;
 document.getElementById("hpgu2_intro_div").hidden=false;
}

function loadHPGU2() {
 unpackCollisionMask();
 var mapString=atob(B64_MAP);
 var mapBuffer=new ArrayBuffer(mapString.length);
 var mapArray=new Uint8Array(mapBuffer);
 for(var i=0;i<mapString.length;++i){
  mapArray[i]=mapString.charCodeAt(i);
 }
 document.getElementById("hpgu2_intro_div").hidden=true;
 document.getElementById("png_processing_message").innerText=
  LOADING_MESSAGE_CYCLE[0];
 document.getElementById("png_processing_message").dataset.cycle=0;
 document.getElementById("png_processing_message").hidden=false;
 document.getElementById("png_input").hidden=true;
 document.getElementById("png_input_error").hidden=true;
 parsePNGFile(mapBuffer);
}



/* Called during init and if a file fails to load */
function awaitFile() {
 isFileExpected=true;
 document.getElementById("hyopCanvasArea").hidden=true;
 document.getElementById("hyop_min_width").innerText=MIN_PNG_WIDTH;
 document.getElementById("hyop_min_height").innerText=MIN_PNG_HEIGHT;
 document.getElementById("hyop_max_width").innerText=MAX_PNG_SIZE;
 document.getElementById("hyop_max_height").innerText=MAX_PNG_SIZE;
 document.getElementById("png_input_div").hidden=false;
 document.body.ondrop=onFileDrop;
 document.body.ondragover=function(e) {
  if(isFileExpected) {
   e.dataTransfer.dropEffect="copy";
   e.preventDefault();
   document.getElementById("png_dropzone_overlay").hidden=false;
  }
 }
 document.body.ondragleave=function(e) {
  document.getElementById("png_dropzone_overlay").hidden=true;
 }
 document.getElementById("png_input").hidden=false;
 document.getElementById("png_input").onchange=onFileInputChange;
 document.getElementById("png_input_error").hidden=true;
 document.getElementById("png_processing_message").hidden=true;
}

/* Try to load a file that was dropped into the input zone */
function onFileDrop(event) {
 event.preventDefault();
 document.getElementById("png_dropzone_overlay").hidden=true;
 if(isFileExpected && event.dataTransfer.items) {
  for(var i=0;i<event.dataTransfer.items.length;++i) {
   if(event.dataTransfer.items[i].kind==="file") {
    tryToLoadFile(event.dataTransfer.items[i].getAsFile());
    return;
   }
  }
 }
 
 if(event.dataTransfer.files && event.dataTransfer.files.length) {
  tryToLoadFile(event.dataTransfer.files[0]); 
 }
}

/* Try to load a file that was opened via form input */
function onFileInputChange(event) {
 if(!event.target.hidden && event.target.files.length) {
  tryToLoadFile(event.target.files[0]);
 }
}

function tryToLoadFile(file) {
 isFileExpected=false;
 document.getElementById("png_processing_message").innerText=
  LOADING_MESSAGE_CYCLE[0];
 document.getElementById("png_processing_message").dataset.cycle=0;
 document.getElementById("png_processing_message").hidden=false;
 document.getElementById("png_input").hidden=true;
 document.getElementById("png_input_error").hidden=true;
 file.arrayBuffer().then(parsePNGFile,unreadableFileError);
}

/* Called on an input file, attempts to load it as a level */
function parsePNGFile(arrayBuffer) {
 var reader=new PNGReader(arrayBuffer);
 reader.parse((err,png) => {
  if(err) {
   fileInputError("File was not a PNG, or it used "+
		  "a feature that PNG.js could not understand.");
   return;
  }
  loadLevelAsync(png).then(level=> {
   try {
    document.getElementById("png_input_div").hidden=true;
    document.getElementById("png_dropzone_overlay").hidden=true;
    document.getElementById("hyopCanvasArea").hidden=false;
    startGame(level);
   }
   catch(e) {
    console.log(e);
    fileInputError("Unexpected error after loading the level, while "+
		   "trying to start the game: "+
		   e.name+" / "+e.message);
    return;
   }
  }).catch(e => {
   if(e) {
    if(typeof(e)==="string") {
     fileInputError("Unexpected error trying to load the level: "+
		    e);
    }
    else {
     console.log(e)
     fileInputError("Unexpected error trying to load the level: "+
		    e.name+" / "+e.message);
    }
   }
  });
 });
}

const LOADING_MESSAGE_CYCLE=[
 "File is loading, please wait...",
 ".File is loading, please wait..",
 "..File is loading, please wait.",
 "...File is loading, please wait",
 "..File is loading, please wait.",
 ".File is loading, please wait..",
];

function animateLoadingMessage(percent) {
 var element=document.getElementById("png_processing_message");
 var cycle=+element.dataset.cycle;
 cycle=(cycle+1)%LOADING_MESSAGE_CYCLE.length;
 document.getElementById("png_processing_message").dataset.cycle=cycle;
 document.getElementById("png_processing_message").innerText=
  LOADING_MESSAGE_CYCLE[cycle]+" ("+percent+"%)";
}

function unreadableFileError() { fileInputError("Could not open file."); }
function fileInputError(msg) {
 awaitFile();
 document.getElementById("png_input_error").hidden=false;
 document.getElementById("png_input_error").innerText=msg;
}

// reverse of what generate_spritesheet_mask.js does
function unpackCollisionMask() {
 SPRITESHEET_MASK.alpha=new Uint8Array(SPRITESHEET_MASK.width*
				       SPRITESHEET_MASK.height);
 var index=0;
 function write(value) {
  if(index<SPRITESHEET_MASK.alpha.length) {
   SPRITESHEET_MASK.alpha[index]=value;
   ++index;
  }
 }
 var state=false;
 SPRITESHEET_MASK.rle.forEach(n => {
  for(var i=0;i<n;++i) {
   write(state?255:0);
  }
  state=!state;
 });
}
