"use strict"

// This constant is just an engine detail; changing it won't have any visible
// effect
const SCROLL_TILE_SIZE=256;

// How many pixels to process at once before yielding the Javascript thread
// and animating the loading display. Too big and the browser's
// slow-script-detection kicks in; too small and all the yielding slows down
// the loading.
const PIXEL_PROCESS_COUNT=200000;

var registeredStamps=[];

function loadLevelAsync(png) {
 return new Promise((resolve,reject) => {
  // check file size
  if(png.width>MAX_PNG_SIZE) {
   fileInputError("File was too wide. The maximum width is "+MAX_PNG_SIZE+
		  " pixels.");   
   reject();
   return;
  }
  if(png.height>MAX_PNG_SIZE) {
   fileInputError("File was too tall. The maximum height is "+MAX_PNG_SIZE+
		  " pixels.");   
   reject();
   return;
  }
  if(png.width<MIN_PNG_WIDTH) {
   fileInputError("File was too narrow. The minimum width is "+MIN_PNG_WIDTH+
		  " pixels.");
   reject();
   return;
  }
  if(png.height<MIN_PNG_HEIGHT) {
   fileInputError("File was too short. The minimum height is "+MIN_PNG_HEIGHT+
		  " pixels.");
   reject();
   return;
  }  
 
  // Why do we use PNGReader instead of an
  // HTMLImageElement? answer: we need _raw_ RGB values, and HTMLImageElement
  // applies colorspace corrections and possibly other GPU weirdness.
  var pixelData=png.getRGBA8Array();
  
  // turn the raw decoded RGBA into what the game actually needs,
  // with stamps erased, black pixels turned transparent,
  // and otherwise semitransparent pixels turned solid
  findStampsAndEraseAsync(png,pixelData).then( (stamps)=> {
   normalizeBackground(pixelData);
   
   // where do we start the game? on a start-point stamp if we've got one,
   // or otherwise some open space chosen arbitrarily
   var startPoint=getStartPoint(png,pixelData,stamps);
   if(!startPoint) {
    fileInputError("File contains no marked spawn point and no "+
		   "full-black or full-transparent space, "+
		   "so there is nowhere to spawn the player.");
    reject();
    return;
   }
   
   // also split the big array into many small ImageData objects
   // that we can page for scrolling; note that this is _after_ we've
   // erased stamps and normalized background
   var tiles=splitToTiles(png,pixelData);

   function countModifiers(name) {
    var count=0;
    var plus=name+"+";
    if(plus in stamps.bySpecialType) {
     count+=stamps.bySpecialType[plus].length;
    }
    var minus=name+"-";
    if(minus in stamps.bySpecialType) {
     count-=stamps.bySpecialType[minus].length;
    }
    return count;
   }
   
   var level={
    width:png.width,
    height:png.height,
    pixelData: pixelData,
    stamps: stamps,
    tiles: tiles,
    startX: startPoint.x,
    startY: startPoint.y,
    gravityModifier: countModifiers("gravity"),
    fuelModifier: countModifiers("fuel"),
    armorModifier: countModifiers("armor"),
    livesModifier: countModifiers("lives"),
   }
   resolve(level);
  }); // end of findStampsAndEraseAsync.then
 }); // end of entire loadLevelAsync promise
}

/* This checks the templates in registeredStamps whenever it sees a
   magenta pixel, and erases matches, resolving a promise with the
   stamps. It only processes PIXEL_PROCESS_COUNT pixels at a time to avoid
   slow-script detection and keep the loading screen text animating.
*/
function findStampsAndEraseAsync(png,pixelData) {
 return new Promise((resolve,reject) => {
  var iterator=findStampsAndEraseGenerator(png,pixelData,PIXEL_PROCESS_COUNT);
  function executeStep() {
   var outcome=iterator.next();
   if(outcome.done) {
    // final value is stamps
    resolve(outcome.value);
   }
   else {
    // intermediate value is percentage progress to send to loading message
    animateLoadingMessage(outcome.value);
    setTimeout(executeStep,0);
   }
  }
  executeStep();
 })
}

// This is written as a generator so the for loop can be periodically
// interrupted by yield statements without needing to wrap steps of it in
// extra promise layers. It yields a progress percentage, then returns
// the stamps.
function* findStampsAndEraseGenerator(png,pixelData,pixelsPerYield) {
 var pixelCount=0, progressCount=0, progressMax=png.width*png.height;
 var stamps={
  bySpecialType:{},
  asList:[]  
 }
 for(var y=0;y<png.height;++y) {
  for(var x=0;x<png.width;++x) {
   var index=y*png.width+x;
   if(pixelData[index*4]==255 &&
      pixelData[index*4+1]==0 &&
      pixelData[index*4+2]==255 &&
      pixelData[index*4+3]==255) {
    //x,y is a magenta pixel, maybe part of a stamp
    for(var stampIdx in registeredStamps) {
     var regStamp=registeredStamps[stampIdx];
     if(testStampMatch(png,pixelData,x,y,regStamp)) {
      var stamp={
       registeredStamp:regStamp,
       left:x-regStamp.firstPixelXOffset,
       top:y-regStamp.firstPixelYOffset
      }
      // erase the magenta pixels that are part of the stamp!
      eraseStamp(png,pixelData,stamp);
      stamps.asList.push(stamp);
      if(regStamp.specialType) {
       if(!(regStamp.specialType in stamps.bySpecialType)) {
 	stamps.bySpecialType[regStamp.specialType]=[]
       }
       stamps.bySpecialType[regStamp.specialType].push(stamp);
      }
     }
    }
   }
   ++pixelCount;
   ++progressCount;
   if(pixelCount>=pixelsPerYield) {
    pixelCount=0;
    yield Math.floor(progressCount*100/progressMax);
   }
  }
 }
 return stamps;
}


/* This tests one template in registeredStamps against a chunk of
   pixel data. Some pixels have to be magenta, and some pixels have
   to be either fully transparent or a color other than magenta. */
function testStampMatch(png,pixelData,firstPixelX,firstPixelY,regStamp) {
 var stampLeft=firstPixelX-regStamp.firstPixelXOffset;
 var stampTop=firstPixelY-regStamp.firstPixelYOffset;
 for(var yInTemplate=0;
     yInTemplate<regStamp.matchTemplate.length;
     ++yInTemplate) {
  for(var xInTemplate=0;
      xInTemplate<regStamp.matchTemplate[yInTemplate].length;
      ++xInTemplate) {
   var toMatch=regStamp.matchTemplate[yInTemplate][xInTemplate];
   if(toMatch) {
    var isMagenta;
    var xInData=stampLeft+xInTemplate+regStamp.templateXOffset;
    var yInData=stampTop+yInTemplate+regStamp.templateYOffset;
    if(xInData<0 || xInData>=png.width || yInData<0 || yInData>=png.height) {
     isMagenta=false;
    }
    else {
     var index=yInData*png.width+xInData;
     isMagenta=(pixelData[index*4]==255 &&
		pixelData[index*4+1]==0 &&
		pixelData[index*4+2]==255 &&
		pixelData[index*4+3]>0);
    }
    if(toMatch==-1 && isMagenta) { return false; }
    if(toMatch==1 && !isMagenta) { return false; }
   }
  }
 }
 return true;
}

/* This erases a stamp's magenta pixels. It's only called after a
   corresponding testStampMatch has found those pixels. */
function eraseStamp(png,pixelData,stamp) {
 var regStamp=stamp.registeredStamp;
 for(var yInTemplate=0;
     yInTemplate<regStamp.matchTemplate.length;
     ++yInTemplate) {
  for(var xInTemplate=0;
      xInTemplate<regStamp.matchTemplate[yInTemplate].length;
      ++xInTemplate) {
   var toMatch=regStamp.matchTemplate[yInTemplate][xInTemplate];
   if(toMatch==1) {
    var xInData=stamp.left+xInTemplate+regStamp.templateXOffset;
    var yInData=stamp.top+yInTemplate+regStamp.templateYOffset;
    if(xInData<0 || xInData>=png.width || yInData<0 || yInData>=png.height) {
     console.warn("Trying to erase an out-of-bounds pixel somehow");
    }
    else {
     var index=yInData*png.width+xInData;
     pixelData[index*4+3]=0;
    }
   }
  }
 }
 return true;
}

/* This makes the open space alpha-0 and the non-open space alpha-255. */
function normalizeBackground(pixelData) {
 for(var i=0;i<pixelData.length;i++) {
  var a=pixelData[i*4+3];
  if(a>0) {
   var r=pixelData[i*4];
   var g=pixelData[i*4+1];
   var b=pixelData[i*4+2];
   if(r==0 && g==0 && b==0) {
    // erase black
    pixelData[i*4+3]=0;
   }
   else if(a!=255) {
    // turn semitransparent to solid if it wasn't black
    pixelData[i*4+3]=255;
   }
  }
 }
}

/* There should only be one start point defined, either with a start
   or startcheckpoint stamp, but if there's ambiguity we can fall back to
   more logic too, just so something will play. */
function getStartPoint(png,pixelData,stamps) {
 var startStamp;
 if(stamps.bySpecialType["start"]) {
  startStamp=stamps.bySpecialType["start"][0];
 }
 else if(stamps.bySpecialType["startcheckpoint"]) {
  startStamp=stamps.bySpecialType["startcheckpoint"][0];
 }
 else if(stamps.bySpecialType["checkpoint"]) {
  // choice of which checkpoint is arbitrary, but an arbitrary
  // checkpoint is less arbitrary than an arbitrary pixel
  startStamp=stamps.bySpecialType["checkpoint"][0];
 }
 if(startStamp) {
  return {
   x:startStamp.left+Math.floor(startStamp.registeredStamp.width/2),
   y:startStamp.top+Math.floor(startStamp.registeredStamp.height/2),
  }
 }
 else {
  // pick an arbitrary open pixel
  for(var y=0;y<png.height;++y) {
   for(var x=0;x<png.width;++x) {
    var i=y*png.width+x;
    if(pixelData[i*4+3]==0) {
     return {
      x:x,
      y:y
     }
    }
   }
  }
 }
 // there is nowhere to start!
 return null;
}

/* Copy the big pixelData into a bunch of ImageData objects so we can
   draw a scrolling window to a canvas. */
function splitToTiles(png, pixelData) {
 var tiles=[]
 for(var tileTop=0;tileTop<png.height;tileTop+=SCROLL_TILE_SIZE) {
  var tileRow=[];
  tiles.push(tileRow);
  for(var tileLeft=0;tileLeft<png.width;tileLeft+=SCROLL_TILE_SIZE) {
   var width=Math.min(png.width-tileLeft,SCROLL_TILE_SIZE);
   var height=Math.min(png.height-tileTop,SCROLL_TILE_SIZE);    
   var tile=new ImageData(width,height);
   tileRow.push(tile);
   for(var yInTile=0;yInTile<height;++yInTile) {
    for(var xInTile=0;xInTile<width;++xInTile) {
     var indexInTile=yInTile*width+xInTile;
     var xInData=xInTile+tileLeft, yInData=yInTile+tileTop;
     var indexInData=yInData*png.width+xInData;
     tile.data[indexInTile*4]=pixelData[indexInData*4]; //r
     tile.data[indexInTile*4+1]=pixelData[indexInData*4+1]; //g
     tile.data[indexInTile*4+2]=pixelData[indexInData*4+2]; //b
     tile.data[indexInTile*4+3]=pixelData[indexInData*4+3]; //a
    }
   }
  }
 }
 return tiles;
}


/* This takes an ascii-art drawing of a stamp, along with the constructor
   for the sprite it spawns, and converts it to
   a registeredStamp data structure. The player spawns differently from
   other sprites, and stamps that might spawn the player use specialType,
   as do stamps that don't actually spawn any sprites.
*/
function registerStamp(drawing,
		       spriteConstructor,
		       specialType) {
 var height=drawing.length, width=1;
 for(var y in drawing) {
  if(drawing[y].length>width) { width=drawing[y].length; }
 }
 var template=[]
 // make template 2 pixels bigger than drawing so there's a 1-pixel margin
 for(var templateY=0;templateY<height+2;++templateY) {
  var templateRow=[]
  template.push(templateRow);
  for(var templateX=0;templateX<width+2;++templateX) {
   templateRow.push(0);
  }
 }
 var firstPixelXOffset=-1;
 var firstPixelYOffset=-1;
 for(var y=0;y<drawing.length;++y) {
  for(var x=0;x<drawing[y].length;++x) {
   if(drawing[y][x]!=" " && drawing[y][x]!=".") {
    // mark this pixel must-be-magenta (even if it was tentatively marked
    // must-not-be-magenta because of a neighbor)
    // Because of the 1-pixel margin, the coordinate in the template
    // is 1 larger than the coordinate in the drawing
    template[y+1][x+1]=1;
    // if the pixels up-left, up, up-right, and left of this one aren't
    // already part of the template, mark them as must-not-be-magenta
    if(!template[y][x]) { template[y][x]=-1; } // UL
    if(!template[y][x+1]) { template[y][x+1]=-1; }  // U    
    if(!template[y][x+2]) { template[y][x+2]=-1; } // UR
    if(!template[y+1][x]) { template[y+1][x]=-1; } // L
    // pixels to the right and below this one haven't been seen yet;
    // mark them must-not-be-magenta for now, but they might turn
    // must-be-magenta later when reached
    template[y+1][x+2]=-1; // R
    template[y+2][x]=-1; // DL
    template[y+2][x+1]=-1; // D
    template[y+2][x+2]=-1; // DR
    if(firstPixelXOffset==-1) {
     // Found the first magenta pixel of the stamp, which is important later
     // when aligning the potential match against a magenta pixel in the image
     firstPixelXOffset=x;
     firstPixelYOffset=y;
    } 
   }
  }
 }
 var regStamp={
  firstPixelXOffset: firstPixelXOffset,
  firstPixelYOffset: firstPixelYOffset,
  matchTemplate:template,
  templateXOffset:-1,
  templateYOffset:-1,
  width:width,
  height:height
 }
 if(specialType) { regStamp.specialType=specialType; }
 if(spriteConstructor) { regStamp.spriteConstructor=spriteConstructor; }
 registeredStamps.push(regStamp);
}

function registerSprite(spriteConstructor) {
 if("stamp" in spriteConstructor) {
  registerStamp(spriteConstructor.stamp, spriteConstructor); 
 }
 else {
  var instance=new spriteConstructor(0,0);
  if("stamp" in instance) {
   registerStamp(stamp, spriteConstructor);
  }
  else {
   throw new Error(
    "registerSprite was called on a class with no stamp property"
   );
  }
 }
}


// Most stamps are defined with a corresponding sprite, but
// non-checkpoint player start and the difficulty-tuners have just
// a specialType to use during the level loading process and no sprite class.
registerStamp(
 [
  ".....#.....",
  ".....#.....",
  ".....#.....",
  ".....#.....",
  "...........",
  "####.#.####",
  "...........",
  ".....#.....",
  ".....#.....",
  ".....#.....",
  ".....#.....",
 ],
 null,
 "start"
);

registerStamp(
 [
  "###############",
  "#.............#",
  "#.###.#.#.....#",
  "#.#...#.#.....#",
  "#.###.#.#.###.#",
  "#.#...#.#.....#",
  "#.#...###.....#",
  "#.............#",
  "###############",
 ],
 null,
 "fuel-"
);

registerStamp(
 [
  "###############",
  "#.............#",
  "#.###.#.#.....#",
  "#.#...#.#.....#",
  "#.###.#.#.###.#",
  "#.#...#.#.....#",
  "#.#...###.....#",
  "#.............#",
  "###############",
 ],
 null,
 "fuel-"
);

registerStamp(
 [
  "###############",
  "#.............#",
  "#.###.#.#.....#",
  "#.#...#.#..#..#",
  "#.###.#.#.###.#",
  "#.#...#.#..#..#",
  "#.#...###.....#",
  "#.............#",
  "###############",
 ],
 null,
 "fuel+"
);

registerStamp(
 [
  "###############",
  "#.............#",
  "#.###.##......#",
  "#.#...#.#.....#",
  "#.#.#.##..###.#",
  "#.#.#.#.#.....#",
  "#.###.#.#.....#",
  "#.............#",
  "###############",
 ],
 null,
 "gravity-"
);

registerStamp(
 [
  "###############",
  "#.............#",
  "#.###.##......#",
  "#.#...#.#..#..#",
  "#.#.#.##..###.#",
  "#.#.#.#.#..#..#",
  "#.###.#.#.....#",
  "#.............#",
  "###############",
 ],
 null,
 "gravity+"
);

registerStamp(
 [
  "###############",
  "#.............#",
  "#..#..##......#",
  "#.#.#.#.#.....#",
  "#.###.##..###.#",
  "#.#.#.#.#.....#",
  "#.#.#.#.#.....#",
  "#.............#",
  "###############",
 ],
 null,
 "armor-"
);

registerStamp(
 [
  "###############",
  "#.............#",
  "#..#..##......#",
  "#.#.#.#.#..#..#",
  "#.###.##..###.#",
  "#.#.#.#.#..#..#",
  "#.#.#.#.#.....#",
  "#.............#",
  "###############",
 ],
 null,
 "armor+"
);

registerStamp(
 [
  "###############",
  "#.............#",
  "#.#...###.....#",
  "#.#....#......#",
  "#.#....#..###.#",
  "#.#....#......#",
  "#.###.###.....#",
  "#.............#",
  "###############",
 ],
 null,
 "lives-"
);


registerStamp(
 [
  "###############",
  "#.............#",
  "#.#...###.....#",
  "#.#....#...#..#",
  "#.#....#..###.#",
  "#.#....#...#..#",
  "#.###.###.....#",
  "#.............#",
  "###############",
 ],
 null,
 "lives+"
);

